@@ -1,6 +1,14 @@
 /**
  * perl-libxml-mm.c
- * $Id: perl-libxml-mm.c,v 1.43 2006/06/22 10:08:58 pajas Exp $
+ * $Id$
+ *
+ * This is free software, you may use it and distribute it under the same terms as
+ * Perl itself.
+ *
+ * Copyright 2001-2003 AxKit.com Ltd., 2002-2006 Christian Glahn, 2006-2009 Petr Pajas
+*/
+
+/*
  *
  * Basic concept:
  * perl varies in the implementation of UTF8 handling. this header (together
@@ -8,18 +16,13 @@
  * the core module inorder to avoid cascades of c pragmas
  */
 
-#ifdef __cplusplus
-extern "C" {
-#endif
-
 #include <stdarg.h>
 #include <stdlib.h>
 
-#include "EXTERN.h"
-#include "perl.h"
-#include "XSUB.h"
+#include "perl-libxml-mm.h"
 
-#include <libxml/parser.h>
+#include "XSUB.h"
+#include "ppport.h"
 #include <libxml/tree.h>
 
 #ifdef XML_LIBXML_GDOME_SUPPORT
@@ -29,24 +32,13 @@ extern "C" {
 
 #endif
 
-#ifdef __cplusplus
-}
-#endif
-
 #include "perl-libxml-sax.h"
 
-#ifdef XS_WARNINGS
-#define xs_warn(string) warn(string) 
-/* #define xs_warn(string) fprintf(stderr, string) */
-#else
-#define xs_warn(string)
-#endif
-
 /**
- * this is a wrapper function that does the type evaluation for the 
+ * this is a wrapper function that does the type evaluation for the
  * node. this makes the code a little more readable in the .XS
- * 
- * the code is not really portable, but i think we'll avoid some 
+ *
+ * the code is not really portable, but i think we'll avoid some
  * memory leak problems that way.
  **/
 const char*
@@ -56,7 +48,7 @@ PmmNodeTypeName( xmlNodePtr elem ){
     if ( elem != NULL ) {
         switch ( elem->type ) {
         case XML_ELEMENT_NODE:
-            name = "XML::LibXML::Element";   
+            name = "XML::LibXML::Element";
             break;
         case XML_TEXT_NODE:
             name = "XML::LibXML::Text";
@@ -68,7 +60,7 @@ PmmNodeTypeName( xmlNodePtr elem ){
             name = "XML::LibXML::CDATASection";
             break;
         case XML_ATTRIBUTE_NODE:
-            name = "XML::LibXML::Attr"; 
+            name = "XML::LibXML::Attr";
             break;
         case XML_DOCUMENT_NODE:
         case XML_HTML_DOCUMENT_NODE:
@@ -96,44 +88,224 @@ PmmNodeTypeName( xmlNodePtr elem ){
 }
 
 /*
- * @node: Reference to the node the structure proxies
- * @owner: libxml defines only the document, but not the node owner
- *         (in case of document fragments, they are not the same!)
- * @count: this is the internal reference count!
- * @encoding: this value is missing in libxml2's doc structure
- *
- * Since XML::LibXML will not know, is a certain node is already
- * defined in the perl layer, it can't shurely tell when a node can be
- * safely be removed from the memory. This structure helps to keep
- * track how intense the nodes of a document are used and will not
- * delete the nodes unless they are not refered from somewhere else.
+ * free a hash table
  */
-struct _ProxyNode {
-    xmlNodePtr node;
-    xmlNodePtr owner;
-    int count;
-    int encoding; 
-};
+void
+PmmFreeHashTable(xmlHashTablePtr table)
+{
+	if( xmlHashSize(table) > 0 ) {
+		warn("PmmFreeHashTable: not empty\n");
+		/* PmmDumpRegistry(table); */
+	}
+	/*	warn("Freeing table %p with %d elements in\n", table, xmlHashSize(table)); */
+	xmlHashFree(table, NULL);
+}
 
-/* helper type for the proxy structure */
-typedef struct _ProxyNode ProxyNode;
+#ifdef XML_LIBXML_THREADS
 
-/* pointer to the proxy structure */
-typedef ProxyNode* ProxyNodePtr;
+/*
+ * registry of all current proxy nodes
+ *
+ * other classes like XML::LibXSLT must get a pointer
+ * to this registry via XML::LibXML::__proxy_registry
+ *
+ */
+extern SV* PROXY_NODE_REGISTRY_MUTEX;
 
-/* this my go only into the header used by the xs */
-#define SvPROXYNODE(x) ((ProxyNodePtr)SvIV(SvRV(x)))
-#define SvNAMESPACE(x) ((xmlNsPtr)SvIV(SvRV(x)))
+/* Utility method used by PmmDumpRegistry */
+void PmmRegistryDumpHashScanner(void * payload, void * data, xmlChar * name)
+{
+	LocalProxyNodePtr lp = (LocalProxyNodePtr) payload;
+	ProxyNodePtr node = (ProxyNodePtr) lp->proxy;
+	const char * CLASS = PmmNodeTypeName( PmmNODE(node) );
+	warn("%s=%p with %d references (%d perl)\n",CLASS,node,PmmREFCNT(node),lp->count);
+}
 
-#define PmmREFCNT(node)      node->count
-#define PmmREFCNT_inc(node)  node->count++
-#define PmmNODE(thenode)     thenode->node
-#define PmmOWNER(node)       node->owner
-#define PmmOWNERPO(node)     ((node && PmmOWNER(node)) ? (ProxyNodePtr)PmmOWNER(node)->_private : node)
+/*
+ * dump the current thread's node registry to STDERR
+ */
+void
+PmmDumpRegistry(xmlHashTablePtr r)
+{
+	if( r )
+	{
+		SvLOCK(PROXY_NODE_REGISTRY_MUTEX);
+		warn("%d total nodes\n", xmlHashSize(r));
+		xmlHashScan(r, PmmRegistryDumpHashScanner, NULL);
+		SvUNLOCK(PROXY_NODE_REGISTRY_MUTEX);
+	}
+}
+
+/*
+ * returns the address of the proxy registry
+ */
+xmlHashTablePtr*
+PmmProxyNodeRegistryPtr(ProxyNodePtr proxy)
+{
+	croak("PmmProxyNodeRegistryPtr: TODO!\n");
+	return NULL;
+	/*   return &PmmREGISTRY; */
+}
+
+/*
+ * efficiently generate a string representation of the given pointer
+ */
+#define _PMM_HASH_NAME_SIZE(n) n+(n>>3)+(n%8>0 ? 1 : 0)
+xmlChar *
+PmmRegistryName(void * ptr)
+{
+	unsigned long int v = (unsigned long int) ptr;
+	int HASH_NAME_SIZE = _PMM_HASH_NAME_SIZE(sizeof(void*));
+	xmlChar * name;
+	int i;
+
+	name = (xmlChar *) safemalloc(HASH_NAME_SIZE+1);
+
+	for(i = 0; i < HASH_NAME_SIZE; ++i)
+	{
+		name[i] = (xmlChar) (128 | v);
+		v >>= 7;
+	}
+	name[HASH_NAME_SIZE] = '\0';
+
+	return name;
+}
+
+/*
+ * allocate and return a new LocalProxyNode structure
+ */
+LocalProxyNodePtr
+PmmNewLocalProxyNode(ProxyNodePtr proxy)
+{
+	LocalProxyNodePtr lp;
+    Newc(0, lp, 1, LocalProxyNode, LocalProxyNode);
+	lp->proxy = proxy;
+	lp->count = 0;
+	return lp;
+}
+
+/*
+ * @proxy: proxy node to register
+ *
+ * adds a proxy node to the proxy node registry
+ */
+LocalProxyNodePtr
+PmmRegisterProxyNode(ProxyNodePtr proxy)
+{
+	xmlChar * name = PmmRegistryName( proxy );
+	LocalProxyNodePtr lp = PmmNewLocalProxyNode( proxy );
+        /* warn("LibXML registers proxy node with %p\n",PmmREGISTRY); */
+	SvLOCK(PROXY_NODE_REGISTRY_MUTEX);
+	if( xmlHashAddEntry(PmmREGISTRY, name, lp) )
+		croak("PmmRegisterProxyNode: error adding node to hash, hash size is %d\n",xmlHashSize(PmmREGISTRY));
+	SvUNLOCK(PROXY_NODE_REGISTRY_MUTEX);
+	Safefree(name);
+	return lp;
+}
+
+/* utility method for PmmUnregisterProxyNode */
+/* PP: originally this was static inline void, but on AIX the compiler
+   did not chew it, so I'm removing the inline */
+static void
+PmmRegistryHashDeallocator(void *payload, xmlChar *name)
+{
+	Safefree((LocalProxyNodePtr) payload);
+}
+
+/*
+ * @proxy: proxy node to remove
+ *
+ * removes a proxy node from the proxy node registry
+ */
+void
+PmmUnregisterProxyNode(ProxyNodePtr proxy)
+{
+	xmlChar * name = PmmRegistryName( proxy );
+        /* warn("LibXML unregistering proxy node with %p\n",PmmREGISTRY); */
+	SvLOCK(PROXY_NODE_REGISTRY_MUTEX);
+	if( xmlHashRemoveEntry(PmmREGISTRY, name, PmmRegistryHashDeallocator) )
+            croak("PmmUnregisterProxyNode: error removing node from hash\n");
+	Safefree(name);
+	SvUNLOCK(PROXY_NODE_REGISTRY_MUTEX);
+}
+
+/*
+ * lookup a LocalProxyNode in the registry
+ */
+LocalProxyNodePtr
+PmmRegistryLookup(ProxyNodePtr proxy)
+{
+	xmlChar * name = PmmRegistryName( proxy );
+	LocalProxyNodePtr lp = xmlHashLookup(PmmREGISTRY, name);
+	Safefree(name);
+	return lp;
+}
+
+/*
+ * increment the local refcount for proxy
+ */
+void
+PmmRegistryREFCNT_inc(ProxyNodePtr proxy)
+{
+  /* warn("Registry inc\n"); */
+	LocalProxyNodePtr lp = PmmRegistryLookup( proxy );
+	if( lp )
+		lp->count++;
+	else
+		PmmRegisterProxyNode( proxy )->count++;
+}
+
+/*
+ * decrement the local refcount for proxy and remove the local pointer if zero
+ */
+void
+PmmRegistryREFCNT_dec(ProxyNodePtr proxy)
+{
+  /* warn("Registry dec\n"); */
+	LocalProxyNodePtr lp = PmmRegistryLookup(proxy);
+	if( lp && --(lp->count) == 0 )
+		PmmUnregisterProxyNode(proxy);
+}
+
+/*
+ * internal, used by PmmCloneProxyNodes
+ */
+void *
+PmmRegistryHashCopier(void *payload, xmlChar *name)
+{
+	ProxyNodePtr proxy = ((LocalProxyNodePtr) payload)->proxy;
+	LocalProxyNodePtr lp;
+	Newc(0, lp, 1, LocalProxyNode, LocalProxyNode);
+	memcpy(lp, payload, sizeof(LocalProxyNode));
+	PmmREFCNT_inc(proxy);
+	return lp;
+}
+
+/*
+ * increments all proxy node counters by one (called on thread spawn)
+ */
+void
+PmmCloneProxyNodes()
+{
+	SV *sv_reg = get_sv("XML::LibXML::__PROXY_NODE_REGISTRY",0);
+	xmlHashTablePtr reg_copy;
+	SvLOCK(PROXY_NODE_REGISTRY_MUTEX);
+	reg_copy = xmlHashCopy(PmmREGISTRY, PmmRegistryHashCopier);
+	SvIV_set(SvRV(sv_reg), PTR2IV(reg_copy));
+	SvUNLOCK(PROXY_NODE_REGISTRY_MUTEX);
+}
+
+/*
+ * returns the current number of proxy nodes in the registry
+ */
+int
+PmmProxyNodeRegistrySize()
+{
+	return xmlHashSize(PmmREGISTRY);
+}
+
+#endif /* end of XML_LIBXML_THREADS */
 
-#define PmmENCODING(node)    node->encoding
-#define PmmNodeEncoding(node) ((ProxyNodePtr)(node->_private))->encoding
-#define PmmDocEncoding(node) (node->charset)
 /* creates a new proxy node from a given node. this function is aware
  * about the fact that a node may already has a proxy structure.
  */
@@ -148,13 +320,24 @@ PmmNewNode(xmlNodePtr node)
     }
 
     if ( node->_private == NULL ) {
-        /* proxy = (ProxyNodePtr)malloc(sizeof(struct _ProxyNode));  */
-        Newc(0, proxy, 1, ProxyNode, ProxyNode);
+        switch ( node->type ) {
+        case XML_DOCUMENT_NODE:
+        case XML_HTML_DOCUMENT_NODE:
+        case XML_DOCB_DOCUMENT_NODE:
+            proxy = (ProxyNodePtr)xmlMalloc(sizeof(struct _DocProxyNode));
+            if (proxy != NULL) {
+                ((DocProxyNodePtr)proxy)->psvi_status = Pmm_NO_PSVI;
+                SetPmmENCODING(proxy, XML_CHAR_ENCODING_NONE);
+            }
+            break;
+        default:
+            proxy = (ProxyNodePtr)xmlMalloc(sizeof(struct _ProxyNode));
+            break;
+        }
         if (proxy != NULL) {
             proxy->node  = node;
             proxy->owner   = NULL;
             proxy->count   = 0;
-            proxy->encoding= 0;
             node->_private = (void*) proxy;
         }
     }
@@ -166,7 +349,7 @@ PmmNewNode(xmlNodePtr node)
 }
 
 ProxyNodePtr
-PmmNewFragment(xmlDocPtr doc) 
+PmmNewFragment(xmlDocPtr doc)
 {
     ProxyNodePtr retval = NULL;
     xmlNodePtr frag = NULL;
@@ -190,12 +373,12 @@ PmmNewFragment(xmlDocPtr doc)
     return retval;
 }
 
-/* frees the node if nessecary. this method is aware, that libxml2
- * has several diffrent nodetypes.
+/* frees the node if necessary. this method is aware that libxml2
+ * has several different nodetypes.
  */
 void
 PmmFreeNode( xmlNodePtr node )
-{  
+{
     switch( node->type ) {
     case XML_DOCUMENT_NODE:
     case XML_HTML_DOCUMENT_NODE:
@@ -212,7 +395,7 @@ PmmFreeNode( xmlNodePtr node )
         break;
     case XML_DTD_NODE:
         if ( node->doc != NULL ) {
-            if ( node->doc->extSubset != (xmlDtdPtr)node 
+            if ( node->doc->extSubset != (xmlDtdPtr)node
                  && node->doc->intSubset != (xmlDtdPtr)node ) {
                 xs_warn( "PmmFreeNode: XML_DTD_NODE\n");
                 node->doc = NULL;
@@ -238,17 +421,17 @@ PmmFreeNode( xmlNodePtr node )
    the subtree if it is not required any more.
  */
 int
-PmmREFCNT_dec( ProxyNodePtr node ) 
-{ 
+PmmREFCNT_dec( ProxyNodePtr node )
+{
     xmlNodePtr libnode = NULL;
-    ProxyNodePtr owner = NULL;  
+    ProxyNodePtr owner = NULL;
     int retval = 0;
 
     if ( node != NULL ) {
         retval = PmmREFCNT(node)--;
 	/* fprintf(stderr, "REFCNT on 0x%08.8X decremented to %d\n", node, PmmREFCNT(node)); */
         if ( PmmREFCNT(node) < 0 )
-            warn( "PmmREFCNT_dec: REFCNT decremented below 0!" );
+            warn( "PmmREFCNT_dec: REFCNT decremented below 0 for %p!", node );
         if ( PmmREFCNT(node) <= 0 ) {
             xs_warn( "PmmREFCNT_dec: NODE DELETION\n" );
 
@@ -280,11 +463,13 @@ PmmREFCNT_dec( ProxyNodePtr node )
             }
             else if ( libnode != NULL ) {
                 xs_warn( "PmmREFCNT_dec:   STANDALONE REAL DELETE\n" );
-                
+
                 PmmFreeNode( libnode );
             }
-            Safefree( node );
-            /* free( node ); */
+			else {
+				xs_warn( "PmmREFCNT_dec:   NO OWNER\n" );
+			}
+            xmlFree( node );
         }
     }
     else {
@@ -307,19 +492,23 @@ PmmREFCNT_dec( ProxyNodePtr node )
  * nodes.
  */
 SV*
-PmmNodeToSv( xmlNodePtr node, ProxyNodePtr owner ) 
+PmmNodeToSv( xmlNodePtr node, ProxyNodePtr owner )
 {
     ProxyNodePtr dfProxy= NULL;
     SV * retval = &PL_sv_undef;
     const char * CLASS = "XML::LibXML::Node";
 
     if ( node != NULL ) {
+#ifdef XML_LIBXML_THREADS
+      if( PmmUSEREGISTRY )
+		SvLOCK(PROXY_NODE_REGISTRY_MUTEX);
+#endif
         /* find out about the class */
         CLASS = PmmNodeTypeName( node );
         xs_warn("PmmNodeToSv: return new perl node of class:\n");
         xs_warn( CLASS );
 
-        if ( node->_private != NULL ) { 
+        if ( node->_private != NULL ) {
             dfProxy = PmmNewNode(node);
             /* fprintf(stderr, " at 0x%08.8X\n", dfProxy); */
         }
@@ -337,13 +526,17 @@ PmmNodeToSv( xmlNodePtr node, ProxyNodePtr owner )
                 }
             }
             else {
-                xs_warn("PmmNodeToSv:   proxy creation failed!\n");
+                croak("XML::LibXML: failed to create a proxy node (out of memory?)\n");
             }
         }
 
         retval = NEWSV(0,0);
         sv_setref_pv( retval, CLASS, (void*)dfProxy );
-        PmmREFCNT_inc(dfProxy); 
+#ifdef XML_LIBXML_THREADS
+	if( PmmUSEREGISTRY )
+	    PmmRegistryREFCNT_inc(dfProxy);
+#endif
+        PmmREFCNT_inc(dfProxy);
         /* fprintf(stderr, "REFCNT incremented on node: 0x%08.8X\n", dfProxy); */
 
         switch ( node->type ) {
@@ -351,12 +544,16 @@ PmmNodeToSv( xmlNodePtr node, ProxyNodePtr owner )
         case XML_HTML_DOCUMENT_NODE:
         case XML_DOCB_DOCUMENT_NODE:
             if ( ((xmlDocPtr)node)->encoding != NULL ) {
-                dfProxy->encoding = (int)xmlParseCharEncoding( (const char*)((xmlDocPtr)node)->encoding );
+                SetPmmENCODING(dfProxy, (int)xmlParseCharEncoding( (const char*)((xmlDocPtr)node)->encoding ));
             }
             break;
         default:
             break;
         }
+#ifdef XML_LIBXML_THREADS
+      if( PmmUSEREGISTRY )
+		SvUNLOCK(PROXY_NODE_REGISTRY_MUTEX);
+#endif
     }
     else {
         xs_warn( "PmmNodeToSv: no node found!\n" );
@@ -365,66 +562,40 @@ PmmNodeToSv( xmlNodePtr node, ProxyNodePtr owner )
     return retval;
 }
 
-/* This is a little helper, that allows us to set the encoding attr. 
- * after broken transformations 
- * 
- * PP: This function is not used!
- */
-void
-PmmFixProxyEncoding( ProxyNodePtr dfProxy ) 
-{
-    xmlNodePtr node = PmmNODE( dfProxy );
-    
-    if ( node != NULL ) {
-        switch ( node->type ) {
-        case XML_DOCUMENT_NODE:
-        case XML_HTML_DOCUMENT_NODE:
-        case XML_DOCB_DOCUMENT_NODE:
-            if ( ((xmlDocPtr)node)->encoding != NULL ) {
-                dfProxy->encoding = (int)xmlParseCharEncoding( (const char*)((xmlDocPtr)node)->encoding );
-            }
-            break;
-        default:
-            dfProxy->encoding = 1;
-            break;
-        }
-    }
-
-}
 
 xmlNodePtr
 PmmCloneNode( xmlNodePtr node, int recursive )
 {
     xmlNodePtr retval = NULL;
-    
+
     if ( node != NULL ) {
         switch ( node->type ) {
         case XML_ELEMENT_NODE:
-		case XML_TEXT_NODE:
-		case XML_CDATA_SECTION_NODE:
-		case XML_ENTITY_REF_NODE:
-		case XML_PI_NODE:
-		case XML_COMMENT_NODE:
-		case XML_DOCUMENT_FRAG_NODE:
-		case XML_ENTITY_DECL: 
-            retval = xmlCopyNode( node, recursive );
-            break;
-		case XML_ATTRIBUTE_NODE:
-            retval = (xmlNodePtr) xmlCopyProp( NULL, (xmlAttrPtr) node );
-            break;
+	case XML_TEXT_NODE:
+	case XML_CDATA_SECTION_NODE:
+	case XML_ENTITY_REF_NODE:
+	case XML_PI_NODE:
+	case XML_COMMENT_NODE:
+	case XML_DOCUMENT_FRAG_NODE:
+	case XML_ENTITY_DECL:
+	  retval = xmlCopyNode( node, recursive ? 1 : 2 );
+	  break;
+	case XML_ATTRIBUTE_NODE:
+	  retval = (xmlNodePtr) xmlCopyProp( NULL, (xmlAttrPtr) node );
+	  break;
         case XML_DOCUMENT_NODE:
-		case XML_HTML_DOCUMENT_NODE:
-            retval = (xmlNodePtr) xmlCopyDoc( (xmlDocPtr)node, recursive );
-            break;
+	case XML_HTML_DOCUMENT_NODE:
+	  retval = (xmlNodePtr) xmlCopyDoc( (xmlDocPtr)node, recursive );
+	  break;
         case XML_DOCUMENT_TYPE_NODE:
         case XML_DTD_NODE:
-            retval = (xmlNodePtr) xmlCopyDtd( (xmlDtdPtr)node );
-            break;
+	  retval = (xmlNodePtr) xmlCopyDtd( (xmlDtdPtr)node );
+	  break;
         case XML_NAMESPACE_DECL:
-            retval = ( xmlNodePtr ) xmlCopyNamespace( (xmlNsPtr) node );
-            break;
+	  retval = ( xmlNodePtr ) xmlCopyNamespace( (xmlNsPtr) node );
+	  break;
         default:
-            break;
+	  break;
         }
     }
 
@@ -435,7 +606,7 @@ PmmCloneNode( xmlNodePtr node, int recursive )
  */
 
 xmlNodePtr
-PmmSvNodeExt( SV* perlnode, int copy ) 
+PmmSvNodeExt( SV* perlnode, int copy )
 {
     xmlNodePtr retval = NULL;
     ProxyNodePtr proxy = NULL;
@@ -464,7 +635,7 @@ PmmSvNodeExt( SV* perlnode, int copy )
         else if ( sv_derived_from( perlnode, "XML::GDOME::Node" ) ) {
             GdomeNode* gnode = (GdomeNode*)SvIV((SV*)SvRV( perlnode ));
             if ( gnode == NULL ) {
-                warn( "no XML::GDOME data found (datastructure empty)" );    
+                warn( "no XML::GDOME data found (datastructure empty)" );
             }
             else {
                 retval = gdome_xml_n_get_xmlNode( gnode );
@@ -485,7 +656,7 @@ PmmSvNodeExt( SV* perlnode, int copy )
 /* extracts the libxml2 owner node from a perl reference
  */
 xmlNodePtr
-PmmSvOwner( SV* perlnode ) 
+PmmSvOwner( SV* perlnode )
 {
     xmlNodePtr retval = NULL;
     if ( perlnode != NULL
@@ -499,10 +670,10 @@ PmmSvOwner( SV* perlnode )
 /* reverse to PmmSvOwner(). sets the owner of the current node. this
  * will increase the proxy count of the owner.
  */
-SV* 
+SV*
 PmmSetSvOwner( SV* perlnode, SV* extra )
 {
-    if ( perlnode != NULL && perlnode != &PL_sv_undef ) {        
+    if ( perlnode != NULL && perlnode != &PL_sv_undef ) {
         PmmOWNER( SvPROXYNODE(perlnode)) = PmmNODE( SvPROXYNODE(extra) );
         PmmREFCNT_inc( SvPROXYNODE(extra) );
         /* fprintf(stderr, "REFCNT incremented on new owner: 0x%08.8X\n", SvPROXYNODE(extra)); */
@@ -516,18 +687,18 @@ void PmmFixOwnerList( xmlNodePtr list, ProxyNodePtr parent );
  * this functions fixes the reference counts for an entire subtree.
  * it is very important to fix an entire subtree after node operations
  * where the documents or the owner node may get changed. this method is
- * aware about nodes that already belong to a certain owner node. 
+ * aware about nodes that already belong to a certain owner node.
  *
  * the method uses the internal methods PmmFixNode and PmmChildNodes to
  * do the real updates.
- * 
- * in the worst case this traverses the subtree twice durig a node 
+ *
+ * in the worst case this traverses the subtree twice during a node
  * operation. this case is only given when the node has to be
- * adopted by the document. Since the ownerdocument and the effective 
+ * adopted by the document. Since the ownerdocument and the effective
  * owner may differ this double traversing makes sense.
- */ 
+ */
 int
-PmmFixOwner( ProxyNodePtr nodetofix, ProxyNodePtr parent ) 
+PmmFixOwner( ProxyNodePtr nodetofix, ProxyNodePtr parent )
 {
     ProxyNodePtr oldParent = NULL;
 
@@ -547,7 +718,7 @@ PmmFixOwner( ProxyNodePtr nodetofix, ProxyNodePtr parent )
         if ( PmmOWNER(nodetofix) != NULL ) {
             oldParent = PmmOWNERPO(nodetofix);
         }
-        
+
         /* The owner data is only fixed if the node is neither a
          * fragment nor a document. Also no update will happen if
          * the node is already his owner or the owner has not
@@ -564,11 +735,12 @@ PmmFixOwner( ProxyNodePtr nodetofix, ProxyNodePtr parent )
             else {
                 PmmOWNER(nodetofix) = NULL;
             }
-            
+
             if ( oldParent != NULL && oldParent != nodetofix )
                 PmmREFCNT_dec(oldParent);
-            
+
             if ( PmmNODE(nodetofix)->type != XML_ATTRIBUTE_NODE
+                 && PmmNODE(nodetofix)->type != XML_DTD_NODE
                  && PmmNODE(nodetofix)->properties != NULL ) {
                 PmmFixOwnerList( (xmlNodePtr)PmmNODE(nodetofix)->properties,
                                  parent );
@@ -629,14 +801,14 @@ PmmFixOwnerNode( xmlNodePtr node, ProxyNodePtr parent )
     if ( node != NULL && parent != NULL ) {
         if ( node->_private != NULL ) {
             xs_warn( "PmmFixOwnerNode: calling PmmFixOwner\n" );
-            PmmFixOwner((ProxyNode *)(node->_private), parent );
+            PmmFixOwner( (ProxyNodePtr)(node->_private), parent );
         }
         else {
             xs_warn( "PmmFixOwnerNode: calling PmmFixOwnerList\n" );
             PmmFixOwnerList(node->children, parent );
-        } 
+        }
     }
-} 
+}
 
 ProxyNodePtr
 PmmNewContext(xmlParserCtxtPtr node)
@@ -654,15 +826,15 @@ PmmNewContext(xmlParserCtxtPtr node)
     }
     return proxy;
 }
- 
+
 int
-PmmContextREFCNT_dec( ProxyNodePtr node ) 
-{ 
+PmmContextREFCNT_dec( ProxyNodePtr node )
+{
     xmlParserCtxtPtr libnode = NULL;
     int retval = 0;
     if ( node != NULL ) {
         retval = PmmREFCNT(node)--;
-	/* fprintf(stderr, "REFCNT on context 0x%08.8X decremented to %d\n", node, PmmREFCNT(node)); */
+	/* fprintf(stderr, "REFCNT on context %p decremented to %d\n", node, PmmREFCNT(node)); */
         if ( PmmREFCNT(node) <= 0 ) {
             xs_warn( "PmmContextREFCNT_dec: NODE DELETION\n" );
             libnode = (xmlParserCtxtPtr)PmmNODE( node );
@@ -697,9 +869,9 @@ PmmContextSv( xmlParserCtxtPtr ctxt )
 
         retval = NEWSV(0,0);
         sv_setref_pv( retval, CLASS, (void*)dfProxy );
-        PmmREFCNT_inc(dfProxy); 
+        PmmREFCNT_inc(dfProxy);
         /* fprintf(stderr, "REFCNT incremented on new context: 0x%08.8X\n", dfProxy); */
-    }         
+    }
     else {
         xs_warn( "PmmContextSv: no node found!\n" );
     }
@@ -708,7 +880,7 @@ PmmContextSv( xmlParserCtxtPtr ctxt )
 }
 
 xmlParserCtxtPtr
-PmmSvContext( SV * scalar ) 
+PmmSvContext( SV * scalar )
 {
     xmlParserCtxtPtr retval = NULL;
 
@@ -739,16 +911,53 @@ PmmSvContext( SV * scalar )
 xmlChar*
 PmmFastEncodeString( int charset,
                      const xmlChar *string,
-                     const xmlChar *encoding ) 
+                     const xmlChar *encoding,
+                     STRLEN len )
 {
     xmlCharEncodingHandlerPtr coder = NULL;
     xmlChar *retval = NULL;
     xmlBufferPtr in = NULL, out = NULL;
 
+    int i;
+    /* first check that the input is not ascii */
+    /* since we do not want to recode ascii as, say, UTF-16 */
+    if (len == 0)
+        len=xmlStrlen(string);
+    for (i=0; i<len; i++) {
+        if(!string[i] || string[i] & 0x80) {
+            break;
+        }
+    }
+    if (i>=len) return xmlStrdup( string );
+    xs_warn("PmmFastEncodeString: string is non-ascii\n");
+
+    if ( charset == XML_CHAR_ENCODING_ERROR){
+        if (xmlStrcmp(encoding,(const xmlChar*)"UTF-16LE")==0) {
+            charset = XML_CHAR_ENCODING_UTF16LE;
+        } else if (xmlStrcmp(encoding,(const xmlChar*) "UTF-16BE")==0) {
+            charset = XML_CHAR_ENCODING_UTF16BE;
+        }
+    }
     if ( charset == XML_CHAR_ENCODING_UTF8 ) {
         /* warn("use UTF8 for encoding ... %s ", string); */
         return xmlStrdup( string );
     }
+    else if ( charset == XML_CHAR_ENCODING_UTF16LE || charset == XML_CHAR_ENCODING_UTF16BE ){
+        /* detect and strip BOM, if any */
+        if (len>=2 && (char)string[0]=='\xFE' && (char)string[1]=='\xFF') {
+            xs_warn("detected BE BOM\n");
+            string += 2;
+            len    -= 2;
+            coder = xmlGetCharEncodingHandler( XML_CHAR_ENCODING_UTF16BE );
+        } else if (len>=2 && (char)string[0]=='\xFF' && (char)string[1]=='\xFE') {
+            xs_warn("detected LE BOM\n");
+            string += 2;
+            len    -= 2;
+            coder = xmlGetCharEncodingHandler( XML_CHAR_ENCODING_UTF16LE );
+        } else {
+	  coder= xmlGetCharEncodingHandler( (xmlCharEncoding)charset );
+        }
+    }
     else if ( charset == XML_CHAR_ENCODING_ERROR ){
         /* warn("no standard encoding %s\n", encoding); */
         coder =xmlFindCharEncodingHandler( (const char *)encoding );
@@ -758,14 +967,13 @@ PmmFastEncodeString( int charset,
     }
     else {
         /* warn( "use document encoding %s (%d)", encoding, charset ); */
-      coder= xmlGetCharEncodingHandler((xmlCharEncoding)charset);
+      coder= xmlGetCharEncodingHandler( (xmlCharEncoding)charset );
     }
 
     if ( coder != NULL ) {
         xs_warn("PmmFastEncodeString: coding machine found \n");
-        in    = xmlBufferCreate();
+        in    = xmlBufferCreateStatic((void*)string, len);
         out   = xmlBufferCreate();
-        xmlBufferCCat( in, (const char *) string );
         if ( xmlCharEncInFunc( coder, out, in ) >= 0 ) {
             retval = xmlStrdup( out->content );
             /* warn( "encoded string is %s" , retval); */
@@ -773,7 +981,7 @@ PmmFastEncodeString( int charset,
         else {
             /* warn( "b0rked encoiding!\n"); */
         }
-        
+
         xmlBufferFree( in );
         xmlBufferFree( out );
         xmlCharEncCloseFunc( coder );
@@ -784,38 +992,48 @@ PmmFastEncodeString( int charset,
 xmlChar*
 PmmFastDecodeString( int charset,
                      const xmlChar *string,
-                     const xmlChar *encoding) 
+                     const xmlChar *encoding,
+                     STRLEN* len )
 {
     xmlCharEncodingHandlerPtr coder = NULL;
     xmlChar *retval = NULL;
     xmlBufferPtr in = NULL, out = NULL;
+    if (len==NULL) return NULL;
+    *len = 0;
+    if ( charset == XML_CHAR_ENCODING_ERROR){
+        if (xmlStrcmp(encoding,(const xmlChar*)"UTF-16LE")==0) {
+            charset = XML_CHAR_ENCODING_UTF16LE;
+        } else if (xmlStrcmp(encoding,(const xmlChar*) "UTF-16BE")==0) {
+            charset = XML_CHAR_ENCODING_UTF16BE;
+        }
+    }
 
     if ( charset == XML_CHAR_ENCODING_UTF8 ) {
-        return xmlStrdup( string );
+        retval = xmlStrdup( string );
+        *len = xmlStrlen(retval);
     }
     else if ( charset == XML_CHAR_ENCODING_ERROR ){
         coder = xmlFindCharEncodingHandler( (const char *) encoding );
     }
     else if ( charset == XML_CHAR_ENCODING_NONE ){
-        xs_warn("PmmFastDecodeString: no encoding found\n");
+        warn("PmmFastDecodeString: no encoding found\n");
     }
     else {
-      coder= xmlGetCharEncodingHandler((xmlCharEncoding)charset);
+      coder= xmlGetCharEncodingHandler( (xmlCharEncoding)charset );
     }
 
     if ( coder != NULL ) {
         /* warn( "do encoding %s", string ); */
-        in  = xmlBufferCreate();
+        in  = xmlBufferCreateStatic((void*)string,xmlStrlen(string));
         out = xmlBufferCreate();
-        
-        xmlBufferCat( in, string );        
         if ( xmlCharEncOutFunc( coder, out, in ) >= 0 ) {
-	  retval = xmlCharStrndup((const char *)xmlBufferContent(out), xmlBufferLength(out));
+          *len = xmlBufferLength(out);
+	  retval = xmlStrndup(xmlBufferContent(out), *len);
         }
         else {
-            xs_warn("PmmFastEncodeString: decoding error\n");
+            /* xs_warn("PmmFastEncodeString: decoding error\n"); */
         }
-        
+
         xmlBufferFree( in );
         xmlBufferFree( out );
         xmlCharEncCloseFunc( coder );
@@ -823,21 +1041,21 @@ PmmFastDecodeString( int charset,
     return retval;
 }
 
-/** 
+/**
  * encodeString returns an UTF-8 encoded String
  * while the encodig has the name of the encoding of string
- **/ 
+ **/
 xmlChar*
-PmmEncodeString( const char *encoding, const xmlChar *string ){
+PmmEncodeString( const char *encoding, const xmlChar *string, STRLEN len ){
     xmlCharEncoding enc;
     xmlChar *ret = NULL;
-    
+
     if ( string != NULL ) {
         if( encoding != NULL ) {
             xs_warn("PmmEncodeString: encoding to UTF-8 from:\n");
             xs_warn( encoding );
             enc = xmlParseCharEncoding( encoding );
-            ret = PmmFastEncodeString( enc, string, (const xmlChar *)encoding );
+            ret = PmmFastEncodeString( enc, string, (const xmlChar *)encoding,len);
         }
         else {
             /* if utf-8 is requested we do nothing */
@@ -847,59 +1065,32 @@ PmmEncodeString( const char *encoding, const xmlChar *string ){
     return ret;
 }
 
-/**
- * decodeString returns an $encoding encoded string.
- * while string is an UTF-8 encoded string and 
- * encoding is the coding name
- **/
-char*
-PmmDecodeString( const char *encoding, const xmlChar *string){
-    char *ret=NULL;
-    xmlCharEncoding enc;
-
-    if ( string != NULL ) {
-        xs_warn( "PmmDecodeString called\n" );
-        if( encoding != NULL ) {
-            enc = xmlParseCharEncoding( encoding );
-            ret = (char*)PmmFastDecodeString( enc, string, (const xmlChar*)encoding );
-            xs_warn( "PmmDecodeString done\n" );
-        }
-        else {
-            ret = (char*)xmlStrdup(string);
-        }
-    }
-    return ret;
-}
-
-
 SV*
 C2Sv( const xmlChar *string, const xmlChar *encoding )
 {
     SV *retval = &PL_sv_undef;
     xmlCharEncoding enc;
-    STRLEN len = 0;
 
     if ( string != NULL ) {
         if ( encoding != NULL ) {
             enc = xmlParseCharEncoding( (const char*)encoding );
         }
         else {
-            enc = XML_CHAR_ENCODING_NONE;
+	  enc = (xmlCharEncoding)0;
         }
         if ( enc == 0 ) {
             /* this happens if the encoding is "" or NULL */
             enc = XML_CHAR_ENCODING_UTF8;
         }
 
-        len = xmlStrlen( string );
-        retval = newSVpvn( (const char *)string, xmlStrlen(string) );
-   
+        retval = newSVpvn( (const char *)string, (STRLEN) xmlStrlen(string) );
+
         if ( enc == XML_CHAR_ENCODING_UTF8 ) {
-            /* create an UTF8 string. */       
+            /* create an UTF8 string. */
 #ifdef HAVE_UTF8
             xs_warn("C2Sv: set UTF8-SV-flag\n");
             SvUTF8_on(retval);
-#endif            
+#endif
         }
     }
 
@@ -923,10 +1114,10 @@ Sv2C( SV* scalar, const xmlChar *encoding )
             xs_warn( "SV2C:   use UTF8\n" );
             if( !DO_UTF8(scalar) && encoding != NULL ) {
 #else
-            if ( encoding != NULL ) {        
+            if ( encoding != NULL ) {
 #endif
                 xs_warn( "SV2C:   domEncodeString!\n" );
-                ts= PmmEncodeString( (const char *)encoding, string );
+                ts= PmmEncodeString( (const char *)encoding, string, len );
                 xs_warn( "SV2C:   done encoding!\n" );
                 if ( string != NULL ) {
                     xmlFree(string);
@@ -934,7 +1125,7 @@ Sv2C( SV* scalar, const xmlChar *encoding )
                 string=ts;
             }
         }
-             
+
         retval = xmlStrdup(string);
         if (string != NULL ) {
             xmlFree(string);
@@ -961,32 +1152,25 @@ nodeC2Sv( const xmlChar * string,  xmlNodePtr refnode )
                values set by XML::LibXSLT */
 
             if ( PmmNodeEncoding(real_doc) == XML_CHAR_ENCODING_NONE ) {
-                PmmNodeEncoding(real_doc) = XML_CHAR_ENCODING_UTF8;
+                SetPmmNodeEncoding(real_doc, XML_CHAR_ENCODING_UTF8);
             }
 
-            decoded = PmmFastDecodeString( PmmNodeEncoding(real_doc) ,
+            decoded = PmmFastDecodeString( PmmNodeEncoding(real_doc),
                                            (const xmlChar *)string,
-                                           (const xmlChar*)real_doc->encoding);
+                                           (const xmlChar *)real_doc->encoding,
+                                           &len );
 
             xs_warn( "push decoded string into SV" );
-            len = xmlStrlen( decoded );
             retval = newSVpvn( (const char *)decoded, len );
             xmlFree( decoded );
 
             if ( PmmNodeEncoding( real_doc ) == XML_CHAR_ENCODING_UTF8 ) {
-                /* most probably true, since libxml2 always 
-                 * sets doc->charset to UTF8, see tree.c:
-                 *
-                 * The in memory encoding is always UTF8
-                 * This field will never change and would
-                 * be obsolete if not for binary compatibility.
-                 */
 #ifdef HAVE_UTF8
                 xs_warn("nodeC2Sv: set UTF8-SV-flag\n");
                 SvUTF8_on(retval);
-#endif            
+#endif
             }
-           
+
             return retval;
         }
     }
@@ -1012,40 +1196,37 @@ nodeSv2C( SV * scalar, xmlNodePtr refnode )
             if ( scalar != NULL && scalar != &PL_sv_undef ) {
                 STRLEN len = 0;
                 char * t_pv =SvPV(scalar, len);
-                xmlChar* ts = NULL;
-                xmlChar* string = xmlStrdup((xmlChar*)t_pv);
-                if ( xmlStrlen(string) > 0 ) {
+                xmlChar* string = NULL;
+                if ( t_pv && len > 0 ) {
                     xs_warn( "nodeSv2C:   no undefs\n" );
 #ifdef HAVE_UTF8
                     xs_warn( "nodeSv2C:   use UTF8\n" );
-                    if( !DO_UTF8(scalar) && real_dom != NULL && real_dom->encoding != NULL )
-#else
-                    if ( real_dom != NULL && real_dom->encoding != NULL )
+                    if( !DO_UTF8(scalar) ) {
 #endif
-		    {
                         xs_warn( "nodeSv2C:     domEncodeString!\n" );
-                        /*  if ( string == NULL || *string == 0 ) warn("string is empty" ); */
                         /* The following statement is to handle bad
                            values set by XML::LibXSLT */
                         if ( PmmNodeEncoding(real_dom) == XML_CHAR_ENCODING_NONE ) {
-                            PmmNodeEncoding(real_dom) = XML_CHAR_ENCODING_UTF8;
+                            SetPmmNodeEncoding(real_dom, XML_CHAR_ENCODING_UTF8);
                         }
-
-                        ts= PmmFastEncodeString( PmmNodeEncoding(real_dom),
-                                                 string,
-                                                 (const xmlChar*)real_dom->encoding );
+                        /* the following allocates a new string (by xmlStrdup if no conversion is done) */
+                        string= PmmFastEncodeString( PmmNodeEncoding(real_dom),
+                                                 (xmlChar*) t_pv,
+                                                 (const xmlChar*)real_dom->encoding,
+                                                 len);
                         xs_warn( "nodeSv2C:     done!\n" );
-                        if ( string != NULL ) {
-                            xmlFree(string);
-                        }
-                        string=ts;
-                    }
-                    else {
+#ifdef HAVE_UTF8
+                    } else {
                         xs_warn( "nodeSv2C:   no encoding set, use UTF8!\n" );
                     }
+#endif
+                }
+                if (string==NULL) {
+                    return xmlStrndup((xmlChar*)t_pv,len);
+                } else {
+                    return string;
                 }
                 /* if ( string == NULL ) warn( "nodeSv2C:     string is NULL\n" ); */
-                return string;
             }
             else {
                 xs_warn( "nodeSv2C:   return NULL\n" );
@@ -1058,11 +1239,11 @@ nodeSv2C( SV * scalar, xmlNodePtr refnode )
     }
     xs_warn("nodeSv2C: no encoding !!\n");
 
-    return  Sv2C( scalar, NULL ); 
+    return  Sv2C( scalar, NULL );
 }
 
-SV * 
-PmmNodeToGdomeSv( xmlNodePtr node ) 
+SV *
+PmmNodeToGdomeSv( xmlNodePtr node )
 {
     SV * retval = &PL_sv_undef;
 
@@ -1082,34 +1263,34 @@ PmmNodeToGdomeSv( xmlNodePtr node )
                 CLASS = "XML::GDOME::Attr";
                 break;
             case GDOME_TEXT_NODE:
-                CLASS = "XML::GDOME::Text"; 
+                CLASS = "XML::GDOME::Text";
                 break;
             case GDOME_CDATA_SECTION_NODE:
-                CLASS = "XML::GDOME::CDATASection"; 
+                CLASS = "XML::GDOME::CDATASection";
                 break;
             case GDOME_ENTITY_REFERENCE_NODE:
-                CLASS = "XML::GDOME::EntityReference"; 
+                CLASS = "XML::GDOME::EntityReference";
                 break;
             case GDOME_ENTITY_NODE:
-                CLASS = "XML::GDOME::Entity"; 
+                CLASS = "XML::GDOME::Entity";
                 break;
             case GDOME_PROCESSING_INSTRUCTION_NODE:
-                CLASS = "XML::GDOME::ProcessingInstruction"; 
+                CLASS = "XML::GDOME::ProcessingInstruction";
                 break;
             case GDOME_COMMENT_NODE:
-                CLASS = "XML::GDOME::Comment"; 
+                CLASS = "XML::GDOME::Comment";
                 break;
             case GDOME_DOCUMENT_TYPE_NODE:
-                CLASS = "XML::GDOME::DocumentType"; 
+                CLASS = "XML::GDOME::DocumentType";
                 break;
             case GDOME_DOCUMENT_FRAGMENT_NODE:
-                CLASS = "XML::GDOME::DocumentFragment"; 
+                CLASS = "XML::GDOME::DocumentFragment";
                 break;
             case GDOME_NOTATION_NODE:
-                CLASS = "XML::GDOME::Notation"; 
+                CLASS = "XML::GDOME::Notation";
                 break;
             case GDOME_DOCUMENT_NODE:
-                CLASS = "XML::GDOME::Document"; 
+                CLASS = "XML::GDOME::Document";
                 break;
             default:
                 break;
